#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" "$@"

##############################################################################
# csv2vcs [input.csv] [output.vcs]
#
# Generate vCalender formatted output from comma separated input of calender
# events.
# This script can be run on both windows or linux/cygwin. The requirement is
# that tcl/tk is installed and accessible from the machine. Help is displayed
# when wrong number of parameters is specified.
# (WINDOWS) >> tclsh \PATH\csv2vcs.tcl infile.csv outfile.vcs   
# (Linux)   >> /PATH/csv2vcs.tcl infile.csv outfile.vcs
#
# DISCLAIMER 
# You accept all risks and responsibility for losses, damages, costs and other
# consequences resulting directly or indirectly from using this application and
# any information or material retrieved from it.
# COPYRIGHT
# Peter Ljung (c) 2004.
# This application is free for non-commercial use.
# Please contact Peter Ljung (e-mail: # peter.ljung.inc@telia.com) for
# commercial licensing.
##############################################################################

# Improvements:
# - Handle time zones. Currently local time is assumed
# - Handle CATEGORIES parameter

##### utilities ##############################################################

# Format date in ISO8601 format. Only local time zone is used
# date - date/time string e.g. "1972-05-13 09:15:00"
# return - ISO8601 formatted date e.g. "19720513T0915"
proc toISO8601 {date} {
	clock format [clock scan "$date"] -format "%Y%m%dT%H%M%S"
}

# Format only time part of date
# date - date/time string e.g. "1972-05-13 09:15:00"
# return - formatted time e.g. "09:15:00"
proc toTime {date} {
	clock format [clock scan $date] -format "%H:%M:%S"
}

# Format only date part of date
# date - date/time string e.g. "1972-05-13 09:15:00"
# return - date e.g. "1972-05-13"
proc toDate {date} {
	clock format [clock scan $date] -format "%Y-%m-%d"
}

# Return number of seconds from 00:00:00 to time
# time - time e.g. "09:15:00"
# return - number of seconds from 00:00:00 e.g. "33300"
proc toSeconds {time} {
	expr [clock scan $time]-[clock scan "00:00:00"]
}

# Add time (hh:mm:ss) to date (yyyyy-mm-dd hh:mm:ss)
# date - date/time string e.g. "1972-05-13 09:15:00"
# time - time e.g. "01:15:00"
# return - date/time e.g. "1972-05-13 10:30:00"
proc addTime {date time} {
	clock format [expr [clock scan $date]+[toSeconds [toTime $time]]] -format "%Y-%m-%d %H:%M:%S"
}

# Parse generic date string using pattern description of date.
# Output date can be parsed by e.g. clock function.
# When delimiters exist between parameters in date only order of parameters
# are important in pattern. When no delimiters exist the number of digits is
# also used. Pattern match is therefore very robust.
# date - date to be parsed e.g. "Mn 13/5 1972" or "19720513" 
# pattern - description pattern for date e.g. "D/M Y" or "YYYYMMDD" 
# return - date e.g. "1972-05-13"
proc parseDate {date pattern} {
	if {[string length $pattern] == 0} {
		return $date	;# do no formatting	
	}
	# get Y,M,D strings including delimiters between them (possibly empty)
	regexp {[^YMD]*([Y]+|[M]+|[D]+)([^YMD]*)([Y]+|[M]+|[D]+)([^YMD]*)([Y]+|[M]+|[D]+)} $pattern match p1 p2 p3 p4 p5
	if {([string length $p2] > 0) && ([string length $p4] > 0)} {
		# delimiter between Y,M,D. Delimiters are used
		regexp {[^0-9]*([0-9]+)[^0-9]*([0-9]+)[^0-9]*([0-9]+)} $date match s1 s2 s3
	} else {
		# no delimiter between Y,M,D. Number of letters (e.g. YYYY=4) is used
		regsub -all {.} $p1 {[0-9]} p1_exp
		regsub -all {.} $p3 {[0-9]} p3_exp
		regsub -all {.} $p5 {[0-9]} p5_exp
		append p {[^0-9]*} "($p1_exp)" {[^0-9]*} "($p3_exp)" {[^0-9]*} "($p5_exp)"
		regexp $p $date match s1 s2 s3
	}
	# determine order
	array set arr [list $p1 $s1 $p3 $s2 $p5 $s3]
	set year [lindex [array get arr {*Y*}] 1]
	set month [lindex [array get arr {*M*}] 1]
	set day [lindex [array get arr {*D*}] 1]
	set ans "$year-$month-$day"
}

# Parse generic time string using pattern description of date.
# Output date can be parsed by e.g. clock function.
# When delimiters exist between parameters in date only order of parameters
# are important in pattern. When no delimiters exist the number of digits is
# also used. Pattern match is therefore very robust.
# time - date to be parsed e.g. "10.50" or "1115" 
# pattern - description pattern for time e.g. "h.m" or "hhmm" 
# return - time e.g. "10:50"
proc parseTime {time pattern} {
	if {[string length $pattern] == 0} {
		return $time	;# do no formatting	
	}
	# get h,m strings including delimiters between them (possibly empty)
	regexp {[^hm]*([h]+|[m]+)([^hm]*)([h]+|[m]+)} $pattern match p1 p2 p3
	if {[string length $p2] > 0} {
		# delimiter between h,m. Delimiters are used
		regexp {[^0-9]*([0-9]+)[^0-9]*([0-9]+)} $time match s1 s2
	} else {
		# no delimiter between h,m. Number of letters (e.g. hh=2) is used
		regsub -all {.} $p1 {[0-9]} p1_exp
		regsub -all {.} $p3 {[0-9]} p3_exp
		append p {[^0-9]*} "($p1_exp)" {[^0-9]*} "($p3_exp)"
		regexp $p $time match s1 s2
	}
	# determine order
	if {[string match *h* $p1]} {
		set hours $s1; set minutes $s2
	} else {
		set hours $s2; set minutes $s1
	}
	set ans "$hours:$minutes"
}

# Divide line in trimmed list items. Remove empty items.
# str - str to be divided in list
# delim - delimiters chars to divide str by
# return - list of trimmed items
proc toList {str delim} {
	set str [string trim $str "#;"]		
	set tmp [split $str $delim]
	foreach item $tmp {
		if {[string length [string trim $item]] > 0} {
			lappend items [string trim [string trim $item]]
		}
	}
	set items
}

# Initialize global values 
proc initGlobals {} {
	global PATTERN ORDER DEFAULT FORMAT
	array set PATTERN {DATE *DATE* STIME *ST* ETIME *ET* DUR *DUR* LOC *LOC* SUM *SUM* DESC *DESC*}
	array set ORDER   {DATE -1 STIME -1 ETIME -1 DUR -1 LOC -1 SUM -1 DESC -1}
	array set DEFAULT {DATE "" STIME "" ETIME "" DUR "" LOC "" SUM "" DESC ""}
	array set FORMAT  {DATE "" STIME "" ETIME "" DUR "" LOC "" SUM "" DESC ""}
}

# Parse meta info line and configure parsing accordingly.
# Globals ORDER, DEFAULT and FORMAT is set. See docs for details.
# Default values and order can be intermixed.
# str - meta data string e.g. 
#	"## DATE(YY-MM-DD);STIME();SUMMARY;DURATION(hh:mm)=0:30"
proc parseMeta {str} {
	global PATTERN ORDER DEFAULT FORMAT
	initGlobals	;# reset configuration
	set items [toList $str ";\t"]
	set k 0		;# order index
	foreach item $items {
		# get key, format and value
		set key ""; set format ""; set value ""
		set sub_items [toList $item "()="]
		set key [lindex $sub_items 0]
		if {[llength $sub_items] == 3} {
			set format [lindex $sub_items 1]
			set value [lindex $sub_items 2]
		} elseif {[string match *=* $item]} {
			set value [lindex $sub_items 1]
		} else {
			set format [lindex $sub_items 1]
		}
		# identify pattern
		foreach	i [array names PATTERN] { 
			if {[string match $PATTERN($i) $key]} { set p $i; break }
		}
		# set FORMAT, DEFAULT and ORDER
		set FORMAT($p) $format
		set DEFAULT($p) $value
		if {![string match *=* $item]} {
			set ORDER($p) $k
			incr k
		}
	}
}

# Helpers for parseLine
proc getDate {items} {
	global ORDER DEFAULT FORMAT
	parseDate "[lindex $items $ORDER(DATE)]" "$FORMAT(DATE)"
}
proc getTime {items time} {
	global ORDER DEFAULT FORMAT
	parseTime [lindex $items $ORDER($time)] $FORMAT($time)
}
proc getDateTime {items time} {
	set ans "[getDate $items] [getTime $items $time]"
}

# Parse one CSV formatted line and output in vCalender format to file
# str - CSV formatted line separated by ; OR \t
# out_file - opened file where vCalender is written
proc parseLine {str out_file} {
	global PATTERN ORDER DEFAULT FORMAT
	# divide in trimed items
	set items [toList $str ";\t"]	
	# generate vCal output
	puts $out_file "BEGIN:VEVENT"
	puts $out_file "DTSTART:[toISO8601 [getDateTime $items STIME]]"
	if {$ORDER(ETIME) >= 0} {
		# use ETIME column 
		puts $out_file "DTEND:[toISO8601 [getDateTime $items ETIME]]"
	} elseif {$ORDER(DUR) >= 0} {
		# use DUR column
		set end_time [addTime [getDateTime $items STIME] [getTime $items DUR]]
		puts $out_file "DTEND:[toISO8601 $end_time]"
	} else {
		# use default DUR time 
		set end_time [addTime [getDateTime $items STIME] [parseTime $DEFAULT(DUR) $FORMAT(DUR)]]
		puts $out_file "DTEND:[toISO8601 $end_time]"
	}
	# use column for LOC, SUM, DESC if available otherwize use default 
	if {$ORDER(LOC) >= 0} {
		puts $out_file "LOCATION:[lindex $items $ORDER(LOC)]"
	} elseif {[string length $DEFAULT(LOC)] > 0} {
		puts $out_file "LOCATION:$DEFAULT(LOC)"
	}	
	if {$ORDER(SUM) >= 0} {
		puts $out_file "SUMMARY:[lindex $items $ORDER(SUM)]"
	} elseif {[string length $DEFAULT(SUM)] > 0} {
		puts $out_file "SUMMARY:$DEFAULT(SUM)"
	}
	# replace '//' for '=0D=0A=' in DESC
	if {$ORDER(DESC) >= 0} {
		regsub -all {//} [lindex $items $ORDER(DESC)] {=0D=0A=} str
		puts $out_file "DESCRIPTION;QUOTED-PRINTABLE:$str"
	} elseif {[string length $DEFAULT(DESC)] > 0} {
		regsub -all {//} "$DEFAULT(DESC)" {=0D=0A=} str
		puts $out_file "DESCRIPTION;QUOTED-PRINTABLE:$str"
	}	
	puts $out_file "END:VEVENT"
}

# Parse CSV formatted lines from stdin and output in vCalender format.
# Lines beginning with ## is interpreted as meta data. # lines are skipped. 
# in_file - opened file with CSV formatted data
# out_file - opened file where vCalender data is written
proc parse {in_file out_file} {
	puts $out_file "BEGIN:VCALENDAR"
	puts $out_file "PRODID:-//www.lounge.se//csv2vcs//EN"
	puts $out_file "VERSION:1.0"
	puts $out_file "TZ:+01"
	while {[eof $in_file] != 1} {
		set line [gets $in_file]
		switch -regexp -- $line {
		"^##" { parseMeta $line }
		"^#" {}
		default { if {[string length $line] > 5} { parseLine $line $out_file } } 		
		}
	}
	puts $out_file "END:VCALENDAR"
}

#### help text ###############################################################

proc print_help {} {
puts "USAGE: csv2vcs.tcl input.csv output.vcs"
puts "Generate vCalender formatted output from comma separated input of calender events. Output file can be used for import of calender information to organizer programs as MS Outlook or Palm Desktop. See http://www.imc.org/pdi/ for more information on vCalender specification."
puts "The input file shall be formatted with one calender event on each row where each parameter is separated by ; or TAB. Before calender rows a configuration row need to be written which specify how calender rows shall be parsed. An examle input file is shown below."
puts ""
puts "## DATE;STIME;DURATION;LOCATION;SUMMARY;DESCRIPTION"
puts "2004-03-14;16:30;1:15;Kastrup;Flight to London;Flight SK112//Seat 5B"
puts "2004-03-14;19:15;1:05;Kastrup;Flight to Nice;Flight BA312//Seat 7F"
puts ""
puts "Configuration row handles three types of configuration."
puts "1. Order of calender parameters. (MANDATORY)"
puts "2. Formatting information. Currently used for time and dates which aren't written in standard format. (OPTIONAL)"
puts "3. Static value for parameter. (OPTIONAL)"
puts ""
puts "An examle input file which specifies order, formatting and defaults is presented below:"
puts "## DATE(YYMMDD);STIME(hh.mm);SUM;LOC=Cinema(Scala);DUR=2:00"
puts "040314;21.15;Save Nemo"
puts "040321;21.15;Lost in Translation"
puts ""
puts "First three configuration parameters specifies the order of DATE, STIME and SUMMARY parameters. In addition formatting for DATE and STIME is specified within paratheses. The last two configuration parameters species static values for LOCATION and DURATION parameters."
puts ""
puts "Configuration rules:"
puts "* Configuration row begin with ##"
puts "* Comments row begin with #"
puts "* One parameter can only be specified once in the configuration row. Either the order is specified or a default value is specified"
puts "* Several configuration lines is allowed in one file which allow different configuration within the same file. All previous configuration is lost when new configuration line entered"
puts ""
puts "A summary of the supported paramters are presented below:"
puts "NAME         SHORT     MANDATORY    FORMAT     MEANING"
puts "----------------------------------------------------------------"
puts "DATE         DATE      YES          YYYY-MM-DD  start date"
puts "STIME        STIME     YES          hh:mm       start time"
puts "ETIME        ETIME     YES(*)       hh:mm       end time"
puts "DURATION     DUR       YES(*)       hh:mm       duration"
puts "LOCATION     LOC       NO           text        location"
puts "SUMMARY      SUM       YES          text        short description"
puts "DESCRIPTION  DESC      NO           text(**)    detailed description"
puts ""
puts "(*) Either ETIME or DURATION are mandatory (ETIME=STIME+DUR)"
puts "(**) Line break is specified with //"
puts ""
puts "DISCLAIMER"
puts "----------"
puts "You accept all risks and responsibility for losses, damages, costs and other consequences resulting directly or indirectly from using this application and any information or material retrieved from it."
puts ""
puts "COPYRIGHT"
puts "---------"
puts "Peter Ljung (c) 2004."
puts "This application is free for non-commercial use. Please contact Peter Ljung (e-mail: peter.ljung.inc@telia.com) for commercial licensing."
puts ""
puts ">> For comments send an e-mail to peter.ljung.inc@telia.com <<"
}

##### main ###################################################################

if {$argc != 2} {
	puts "Wrong number of arguments!"
	puts ""
	print_help
} else {
	set in_file [open [lindex $argv 0] r]
	set out_file [open [lindex $argv 1] w]
	
	parse $in_file $out_file

	close $in_file
	close $out_file
}

#### end #####################################################################
